unit uoddface_pig;

{$mode delphi}

interface

uses
  Classes, SysUtils,
  uoddface,
  Math,
  xmatdefs,
  matreal,
  uauxiliares,
  uactores,
  ufechas,
  usalasdejuego,
  uDataSetGenerico,
  urosx,
  uInicioYFinal,
  uconstantessimsee;


(*
PLANIFICACION DE INVERSIONES EN GENERACION
==========================================

Se trata de determinar la expansión óptima de unidades de generación de diferentes
teconologías.
Para ello se crea una Sala SimSEE con todas las teconolgías posibles, configurando
los pagos por energía, costo variables y pagos por potencias en forma correcta para
reflejar tanto los costos fijos como variables.

A los efectos de determinar las fechas de ingreso se utilizará el mes-año
mínima. El problema consiste en determinar el cronograma óptimo de decisiones de
inversión.

Sobre cada teconología sobre la que se quiera trabajar, se definierá un registro
con:

"nombre de la máquina dentro de la sala",
"meses_construcción",
"años_vida_util",
"Costo Descisión MUSD/unidad",
"Fecha_Inicio",
"Fecha_Fin",
"MaximoUnidadesPorVez"
"MaximoUnidadesTotalActivas"


"Fecha_Inicio" y "Fecha_Fin" se utilizan como "filtro ventana para las decisiones", poner 0 (cero) si
en los límites que no apliquen y valdrán los de inicio y/o fin de simulación.

Entonces, una decisión (+n, fecha) significa agregar n unidades en la fecha+meses_construccion
y quitar n unidade sen la fecha+meses_construccion+años_vida_util

Por simplicidad, supondremos que la precisión de ingreso de las inversiones es "trimestral"
con lo cual

Para cada teconologia, se crea entonces un vector de entradas de unidades
en base al cual se actua sobre la sala.

*)


type
  TPIG_Tecnologia = class
    Nombre: string; // Nombre de la máquina en SimSEE
    MesesConstruccion: integer; // meses desde la desición a la entrada en servicio
    AniosVidaUtil: integer; // años de vida útil de la unidad
    CostoDesicion: NReal; // MUSD/unidad asignados al momento de toma de desicion
    // como el pago por potencia depende de la disponibilidad
    // puede ser conveniente usar este parámetro en lugar del PP.
    // En definitiva, si la inversión la pagamos igual independientemente
    // de la disponibilidad, este es el lugar de considerarla.
    FechaInicio, FechaFin: TFecha;
    MaximoUnidades: integer; // máximo de unidades que se pueden decidir instalar por
    MaximoUnidadesTotalActivas: integer; // número máximo de unidades activas.
    // esta restricción, depende de la cantidad de unidades activadas
    // con anterioridad. PUede ser útil para reflejar un límite de disponibilidad
    // de combustible.

    FactorUnidadesGeneradoras_: TDAOfNInt;
    // Factor por el que se multiplica las Unidades
    // de inversión para generar las fichas de unidades de generación.
    // por ejemplo, en el caso de de los parques eólicos puede pensarse que
    // son parques de 50 MW las unidades de inversión, pero que las unidades
    // de generación para el SimSEE son molinos de 2MW y entonces el factorUnidadesGeneradoras
    // sería 25.

    class function ReadFromRec(var r: TDataRecord): TPIG_Tecnologia;
    procedure Free;
  end;

  TPIG_Tecnologias = class
    lst: TList;
  private
    function getTecno(index: integer): TPIG_Tecnologia;
    procedure setTecno(index: integer; Value: TPIG_Tecnologia);

  public
    class function ReadFromDB(nid_problema: integer): TPIG_Tecnologias;
    procedure Free;
    property tecno[index: integer]: TPIG_Tecnologia read getTecno write setTecno;
      default;
  end;


  TSimCostos_archi = class
    tasa: NReal; // [p.u.] tasa de descuento en por unidad

    cad_VE: NReal; // [MUSD] valor esperado del CAD
    cad_VaR5pe: NReal; // [MUSD] VaR(5%) del CAD
    cad_VE_aux: NReal; // [MUSD] valor esperado del CAD considerando CF_aux al final
    cad_VaR5pe_aux: NReal; // [MUSD] VaR(5%) del CAD considerando CF_aux al final
    vcad: TDAOfNReal; // [MUSD] vector con los CAD de cada crónica
    vcad_aux: TDAOfNReal;
    // [MUSD] vector de los CAD de cada crónica, considerando CF_aux al final
    vcdp: TDAOfNReal;
    // [MUSD] vector con los costos directos del paso (CDP) acumulados de cada crónica.
    vcffja: TDAOfNReal;
    // [MUSD] vector con el Costo Futuro de Fin de Juego Actualizado de cad crónica.
    vcffja_aux: TDAOfNReal;
    // [MUSD] vector con el Costo Futuro (auxiliar) de Fin de Juego Actualizado de cad crónica.

    constructor CreateRead(archi_res: string; NCronicas: integer);
    procedure Free;

  end;


  TPIG_Problema = class(TProblema)
    CarpetaSala: string;
    NombreSala: string;

    Tecnologias: TPIG_Tecnologias;

    constructor Create(recProblema: TDataRecord; idEjecutor_: integer;
      tmp_rundir_: string);
    class function CreateFromDB(nid_problema: integer; idEjecutor_: integer;
      tmp_rundir_: string): TPIG_Problema;
    function evaluar_(Individuo: TIndividuo; SemillaAleatoria: integer): boolean;
      override;
    procedure Free; override;

    function EtapaTecnoToParametro(kEtapa, kTecno: integer): integer;
    function RunOptSim(sala: string;
      NCronicas, semillaAleatoria_: integer): TSimCostos_archi;

    // Crea la sala y retorna el camino completo a la misma.
    // En la variable MontoInversiones retorna el valor a sumar al CAD
    // para considerar el monto de inversiones al Inicio del a corrida.
    // En la variable EsFactible indica que la sala resultante está dentro de las factibles.
    function CrearSalaSimSEE(individuo: TIndividuo; var MontoInversiones: NReal;
      var EsFactible: boolean): string;

    function BajarCrearSalaSimSEE(nid: integer; var MontoInversiones: NReal): string;

  end;




implementation



function TPIG_Problema.EtapaTecnoToParametro(kEtapa, kTecno: integer): integer;
begin
  Result := kEtapa * tecnologias.lst.Count + kTecno;
end;


constructor TPIG_Problema.Create(recProblema: TDataRecord; idEjecutor_: integer;
  tmp_rundir_: string);

var
  kEtapa: integer;
  kTecno: integer;
  kParametro: integer;
  aTecno: TPIG_Tecnologia;

begin
  inherited Create(recProblema, idEjecutor_, tmp_rundir_);

  // calculamos carpeta de la sala y nombre
  //  CarpetaSala := extractFilePath( ArchiSala );
  CarpetaSala := tmp_rundir_;
  NombreSala := extractFileName(ArchiSala);
  // acomodamos el directorio de la sala
  ArchiSala := tmp_rundir_ + barraDir + NombreSala;
  if (pos('.ese', NombreSala) > 0) then
    Delete(NombreSala, pos('.ese', NombreSala), length(NombreSala) -
      pos('.ese', NombreSala) + 1);


  // cargamos las posibles tecnologías
  tecnologias := TPIG_Tecnologias.ReadFromDB(nid_Problema);
  if tecnologias = nil then
    raise Exception.Create('No pude leer archivo del problema.');

  // Bien, ahora que tenemos las tecnologias definimos los parámetros
  setlength(DescriptoresE, n_etapas * tecnologias.lst.Count);
  setlength(DescriptoresR, 0);
  for kEtapa := 0 to n_etapas - 1 do
    for kTecno := 0 to tecnologias.lst.Count - 1 do
    begin
      kParametro := EtapaTecnoToParametro(kEtapa, kTecno);
      aTecno := tecnologias[kTecno];
      self.DefinirParametroEntero(
        kParametro,
        'e' + IntToStr(kEtapa) + '_' + aTecno.Nombre,
        0, aTecno.MaximoUnidades);
    end;

  CalcularLargosADN;
end;

class function TPIG_Problema.CreateFromDB(nid_problema: integer;
  idEjecutor_: integer; tmp_rundir_: string): TPIG_Problema;
var
  ds: TResultadoQuery;
  r: TDataRecord;
  ap: TPIG_Problema;
begin
  ds := sql_query('SELECT * FROM ofe_problemas WHERE nid = ' + IntToStr(
    nid_problema) + ' LIMIT 1 ');
  r := ds.Next;
  if r <> nil then
    ap := TPIG_Problema.Create(r, idEjecutor_, tmp_rundir_)
  else
    ap := nil;
  ds.Free;
  Result := ap;
end;


function nextpal_sep(var r: string; sep: string): string;
var
  res: string;
  i: integer;
begin
  r := trim(r);
  i := pos(sep, r);
  if i > 0 then
  begin
    res := trim(copy(r, 1, i - 1));
    Delete(r, 1, i);
  end
  else
  begin
    res := r;
    r := '';
  end;
  Result := res;
end;


function nextpal_tab(var r: string): string;
begin
  Result := nextpal_sep(r, #9);
end;




procedure readln_vect(var f: textfile; var v: TDAOfNreal; N: integer);
var
  r: string;
  pal: string;
  k: integer;
begin
  setlength(v, N);
  system.readln(f, r);
  pal := nextpal_tab(r); // caption
  for k := 0 to high(v) do
  begin
    pal := nextpal_tab(r);
    v[k] := StrToFloat(pal);
  end;
end;




class function TPIG_Tecnologia.ReadFromRec(var r: TDataRecord): TPIG_Tecnologia;
var
  res: TPIG_Tecnologia;
  s, pal: string;
  cnt: integer;

begin
  res := TPIG_Tecnologia.Create;
  res.Nombre := r.GetByNameAsString('nombre');
  res.MesesConstruccion := r.GetByNameAsInt('meses_construccion');
  res.AniosVidaUtil := r.GetByNameAsInt('anios_vida_util');
  res.CostoDesicion := r.GetByNameAsFloat('musd_por_unidad');
  res.FechaInicio := TFecha.Create_ISOStr(r.GetByNameAsString('fecha_inicio'));
  res.FechaFin := TFecha.Create_ISOStr(r.GetByNameAsString('fecha_fin'));
  res.MaximoUnidades := r.GetByNameAsInt('max_unid_por_vez');
  res.MaximoUnidadesTotalActivas := r.GetByNameAsInt('max_unid_activas');

  s := r.GetByNameAsString('factor_unidades');
  pal := nextpal_sep(s, ',');
  setlength(res.FactorUnidadesGeneradoras_, 10);
  cnt := 0;
  while pal <> '' do
  begin
    res.FactorUnidadesGeneradoras_[cnt] := StrToInt(pal);
    Inc(cnt);
    pal := nextpal_sep(s, ',');
  end;
  setlength(res.FactorUnidadesGeneradoras_, cnt);

  Result := res;
end;



procedure TPIG_Tecnologia.Free;
begin
  fechaInicio.Free;
  fechaFin.Free;
  inherited Free;
end;



class function TPIG_Tecnologias.ReadFromDB(nid_problema: integer): TPIG_Tecnologias;
var
  aTecno: TPIG_Tecnologia;
  res: TPIG_Tecnologias;
  sql: string;

  ds: TResultadoQuery;
  rec: TDataRecord;

begin
  res := TPIG_Tecnologias.Create;
  res.lst := TList.Create;

  sql := 'SELECT * FROM ofe_PIG_tecnologias WHERE nid_problema = ' +
    IntToStr(NID_Problema) + ' AND activa = 1 ORDER BY nid ';
  ds := sql_query(sql);
  rec := ds.Next;
  while rec <> nil do
  begin
    aTecno := TPIG_Tecnologia.ReadFromRec(rec);
    res.lst.add(aTecno);
    rec := ds.Next;
  end;

  ds.Free;
  Result := res;
end;



procedure TPIG_Tecnologias.Free;
var
  k: integer;
begin
  for k := 0 to lst.Count - 1 do
    tecno[k].Free;
  lst.Free;
  inherited Free;
end;

function TPIG_Tecnologias.getTecno(index: integer): TPIG_Tecnologia;
begin
  Result := lst.items[index];
end;

procedure TPIG_Tecnologias.setTecno(index: integer; Value: TPIG_Tecnologia);
begin
  lst.items[index] := Value;
end;



constructor TSimCostos_archi.CreateRead(archi_res: string; NCronicas: integer);
var
  f: textfile;
begin
  inherited Create;

  Assign(f, archi_res);
  {$I-}
  reset(f);
  {$I+}
  if ioresult <> 0 then
    raise Exception.Create('No encontré archivo: ' + archi_res);

  system.readln(f, tasa);
  system.readln(f, cad_VE);
  system.readln(f, cad_VaR5pe);
  system.readln(f, cad_VE_aux);
  system.readln(f, cad_VaR5pe_aux);

  readln_vect(f, vcad, NCronicas);
  readln_vect(f, vcad_aux, NCronicas);
  readln_vect(f, vcdp, NCronicas);
  readln_vect(f, vcffja, NCronicas);
  readln_vect(f, vcffja_aux, NCronicas);

  closefile(f);
end;

procedure TSimCostos_archi.Free;
begin
  setlength(vcad, 0);
  setlength(vcad_aux, 0);
  setlength(vcdp, 0);
  setlength(vcffja, 0);
  setlength(vcffja_aux, 0);
  inherited Free;
end;




// retorna NIL si falla algo
function TPIG_Problema.RunOptSim(sala: string;
  NCronicas, semillaAleatoria_: integer): TSimCostos_archi;
var
  cmd: string;
  params: array of string;
  archi_simcosto: string;

begin

  if tmp_rundir = '' then
    archi_simcosto := getDir_run + NombreSala + barradir
  else
    archi_simcosto := getDir_run;

  limpiarCarpeta(archi_simcosto, 'simres_*');
  limpiarCarpeta(archi_simcosto, 'simcosto_*');
  limpiarCarpeta(archi_simcosto, 'estado_fin_cron_*');


  setlength(params, 3);
  params[0] := 'sala=' + sala;
  params[1] := 'ejecutor=' + IntToStr(idEjecutor);
  params[2] := 'semilla=' + IntToStr(semillaAleatoria_);

{$IFDEF LINUX}
  cmd := getDir_bin + 'cmdopt';
{$ELSE}
  cmd := getDir_bin + 'cmdopt.exe';
{$ENDIF}
  if not RunChildAndWAIT(cmd, params) then
    raise Exception.Create('no puede correr la cmdopt');


{$IFDEF LINUX}
  cmd := getDir_bin + 'cmdsim';
{$ELSE}
  cmd := getDir_bin + 'cmdsim.exe';
{$ENDIF}
  setlength(params, 4);
  params[0] := 'sala=' + sala;
  params[1] := 'ejecutor=' + IntToStr(idEjecutor);
  params[2] := 'semilla=' + IntToStr(semillaAleatoria_);
  params[3] := 'NCronicasSim=' + IntToStr(NCronicas);

  if not RunChildAndWAIT(cmd, params) then
    raise Exception.Create('no puede correr la cmdsim');
  setlength(params, 0);


  try
    if tmp_rundir = '' then
      archi_simcosto := getDir_run + NombreSala + barradir
    else
      archi_simcosto := getDir_run;

    archi_simcosto := archi_simcosto + barradir + 'simcosto_' + IntToStr(
      semillaAleatoria_) + 'x' + IntToStr(NCronicas) + '_base.xlt';
    Result := TSimCostos_archi.CreateRead(archi_simcosto, NCronicas);
  except
    Result := nil;
    raise Exception.Create('Error al leer archivo de resultados uplanes.425: ' +
      archi_simcosto);
  end;

end;


function TPIG_Problema.CrearSalaSimSEE(individuo: TIndividuo;
  var MontoInversiones: NReal; var EsFactible: boolean): string;
var
  sala: TSalaDejuego;
  delta_unidades: TDAOfNInt;
  kEtapa, kTecno: integer;
  kParametro: integer;
  actor: TActor;
  tecno: TPIG_Tecnologia;
  du: integer;

  afu: TFichaUnidades;
  nUnidades: TDAofNInt;
  Incertidumbres_Alta, Incertidumbres_InicioCronica: TDAOfBoolean;

  total_unidades: integer;
  fecha: TFecha;
  kEtapa_InicioOperacion, kEtapa_FinVidaUtil: integer;
  dt_InicioEtapa: double;
  dt_InicioSim: double;
  qActualizadorDiario: double;

  archiSalaModificada: string;
  MaxUnidadesGeneradorasActivas: integer;
  kTipoUnidad: integer;

begin
  EsFactible := True;
  uInicioYFinal.AlInicio;

  // cargamos la sala de juegos_base

  chdir(carpetaSala);

  sala := TSalaDeJuego.cargarSala(archiSala, '__principal__', True);

  dt_InicioSim := sala.globs.fechaIniSim.dt;
  qActualizadorDiario := power(1 / (1 + sala.globs.TasaDeActualizacion), 1 / 365.2);

  setlength(delta_unidades, n_etapas);



  MontoInversiones := 0;

  for kTecno := 0 to tecnologias.lst.Count - 1 do
  begin
    tecno := tecnologias[kTecno];
    actor := sala.ListaActores.find(tecno.Nombre) as TActor;
    vclear(delta_unidades);

    // Armamos un vector con los Delta_Unidades.
    // para tener entradas y salidas registradas.
    for kEtapa := 0 to n_etapas - 1 do
    begin

      // Chequeamos si la etapa esta en la ventana en que
      // es posible tomar decisiones para esta tecnología.
      dt_InicioEtapa := (fecha_primer_etapa.dt + kEtapa * dias_por_etapa);
      if (dt_InicioEtapa >= tecno.FechaInicio.dt) and
        (dt_InicioEtapa < tecno.FechaFin.dt) then
      begin
        // si estamos en-rango de tomar decisiones antomos entrada y salida.
        kParametro := EtapaTecnoToParametro(kEtapa, kTecno);
        du := Individuo.XE.e(kParametro + 1);
        if du <> 0 then
        begin
          kEtapa_InicioOperacion :=
            kEtapa + trunc((tecno.MesesConstruccion * 365.2 / 12.0) /
            dias_por_etapa + 0.2);
          if (kEtapa_InicioOperacion < n_etapas) then
          begin
            //          MontoInversiones:= MontoInversiones + tecno.CostoDesicion *  power( qActualizadorDiario, dt_InicioEtapa - dt_InicioSim );
            // rch@20121209, Eliana detecta error en calculo de inversiones. Faltaba multiplicar por cantidad de unidades de inversión.
            MontoInversiones :=
              MontoInversiones + tecno.CostoDesicion * du *
              power(qActualizadorDiario, dt_InicioEtapa - dt_InicioSim);

            delta_unidades[kEtapa_InicioOperacion] :=
              delta_unidades[kEtapa_InicioOperacion] + du;

            kEtapa_FinVidaUtil :=
              kEtapa_InicioOperacion + trunc(
              (tecno.AniosVidaUtil * 365.2) / dias_por_etapa + 0.2);

            if (kEtapa_FinVidaUtil < n_etapas) then
              delta_unidades[kEtapa_FinVidaUtil] :=
                delta_unidades[kEtapa_FinVidaUtil] - du;
          end;
        end;
      end;
    end;


    setlength(nUnidades, length(Tecno.FactorUnidadesGeneradoras_));
    setlength(incertidumbres_Alta, length(Tecno.FactorUnidadesGeneradoras_));
    setlength(incertidumbres_InicioCronica, length(Tecno.FactorUnidadesGeneradoras_));

    total_unidades := 0;

    MaxUnidadesGeneradorasActivas := tecno.MaximoUnidadesTotalActivas;

    for kEtapa := 0 to n_etapas - 1 do
    begin
      if delta_unidades[kEtapa] <> 0 then
      begin

        fecha := TFecha.Create_Clone(fecha_primer_etapa);
        fecha.addDias(kEtapa * dias_por_etapa);

        total_unidades := total_unidades + delta_unidades[kEtapa];

        if total_unidades > MaxUnidadesGeneradorasActivas then
          EsFactible := False;

        for kTipoUnidad := 0 to high(nUNidades) do
        begin
          incertidumbres_Alta[kTipoUnidad] := True;
          incertidumbres_InicioCronica[kTipoUnidad] := True;
          nUnidades[kTipoUnidad] :=
            Tecno.FactorUnidadesGeneradoras_[kTIpoUnidad] * total_unidades;
        end;

        afu := TFichaUnidades.Create(0, fecha, nil, nUnidades,
          Incertidumbres_Alta, Incertidumbres_InicioCronica);
        actor.lpdUnidades.insertar(afu);
      end;
    end;
  end;

  archiSalaModificada := getDir_run + barraDir + self.NombreSala + '_oddface_.ese';
  sala.WriteToArchi(archiSalaModificada);
  sala.Free;

  setlength(nUnidades, 0);
  setlength(Incertidumbres_InicioCronica, 0);
  setlength(Incertidumbres_Alta, 0);


  uInicioYFinal.AlFinal;

  Result := archiSalaModificada;
end;


function TPIG_Problema.BajarCrearSalaSimSEE(nid: integer;
  var MontoInversiones: NReal): string;
var
  a: TIndividuo;
  EsFactible: boolean;
begin
  a := LeerIndividuo(nid);
  decodificar_adn(a);
  Result := CrearSalaSimSEE(a, MontoInversiones, EsFactible);
end;

function TPIG_Problema.evaluar_(Individuo: TIndividuo;
  SemillaAleatoria: integer): boolean;
var
  res_SimCostos: TSimCostos_archi;
  archiSalaModificada: string;
  MontoInversiones: double;
  EsFactible: boolean;
begin

  archiSalaModificada := crearSalaSimsee(Individuo, MontoInversiones, EsFactible);
  if EsFactible then
  begin
    res_SimCostos := RunOptSim(archiSalaModificada, NCronicasCronicasPorVez,
      SemillaAleatoria);
    if res_SimCostos <> nil then
    begin
      Individuo.f_histo := TVectR.Create_FromDAofR(res_SimCostos.vcad);
      Individuo.f_histo.MasReal(MontoInversiones);
      Individuo.f_VE := res_SimCostos.cad_VE + MontoInversiones;
      Individuo.f_VaR := Individuo.f_histo.pe_VaR(pe_var);
      Individuo.f_CVaR := Individuo.f_histo.pe_CVaR(pe_var);
      Individuo.f_MIN := individuo.f_histo.e(1);
      Individuo.f_MAX := individuo.f_histo.e(individuo.f_histo.n);
      Individuo.f_Objetivo := ro_VE * Individuo.f_VE + ro_VaR *
        Individuo.f_VaR + ro_CVaR * Individuo.f_CVaR;
      res_SimCostos.Free;
      Result := True;
    end
    else
      raise Exception.Create('Falló la simulación.');
  end
  else
    Result := False;
end;


procedure TPIG_Problema.Free;
begin
  Tecnologias.Free;
  inherited Free;
end;

end.
